#include "api/df_access_control.h"
#include "datafox_library_iv.h"

DF_API_EXTERN_DECORATE int DF_API_CALL DFCEntrance2Identification( int connectionNumber, int deviceAddress, const char *readerTM, const char *identificationId, const char *identificationPin, char *statusText, int *statusTextLength, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || readerTM == nullptr || identificationId == nullptr || identificationPin == nullptr || statusText == nullptr || statusTextLength == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );

        channelBridge->logEnd( "\t%s\t%s\t%s\t%s\t%i\t%i\t(%i, %i)", (readerTM == nullptr) ? "" : readerTM, (identificationId == nullptr) ? "" : identificationId,
            (identificationPin == nullptr) ? "" : identificationPin, (statusTextLength == nullptr) ? 0 : *statusTextLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );

        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );

        channelBridge->logEnd( "\t%s\t%s\t%s\t%s\t%i\t%i\t(%i, %i)", readerTM, identificationId, strlen(identificationPin) ? "*****" : "", statusText, *statusTextLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int iRes = channelBridge->entrance2Identification( l_device_address, readerTM, identificationId, identificationPin, statusText, *statusTextLength );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 6:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_ENTRANCE2_OFF, errorNumber );
                break;

            case 28:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
                break;

            case 29:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
                break;

            default:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%s\t%s\t%s\t%i\t%i\t(%i, %i)", readerTM, identificationId, strlen(identificationPin) ? "*****" : "", statusText, *statusTextLength, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCEntrance2OnlineAction( int connectionNumber, int deviceAddress, const char *readerTM, int mask, int type, int duration, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || readerTM == nullptr || mask == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );

        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t%i\t%i\t(%i, %i)", (readerTM == nullptr) ? "" : readerTM, mask, type, duration, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );

        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t%i\t%i\t(%i, %i)", readerTM, mask, type, duration, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int iRes = channelBridge->entrance2OnlineAction( l_device_address, readerTM, mask, type, duration );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 6:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_ENTRANCE2_OFF, errorNumber );
                break;

            case 28:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
                break;

            case 29:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
                break;

            default:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t0x%08X\t%i\t%i\t%i\t(%i, %i)", (readerTM == nullptr) ? "" : readerTM, mask, type, duration, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCAccessControlIdentification( int connectionNumber, int deviceAddress, int masterId, const char *readerTM, const char *identificationId, const char *identificationPin, char *statusText, int *statusTextLength, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", masterId );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || masterId < 1 || masterId > 9999
        || readerTM == nullptr || identificationId == nullptr || identificationPin == nullptr || statusText == nullptr || statusTextLength == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );

        channelBridge->logEnd( "\t%s\t%s\t%s\t%s\t%i\t%i\t(%i, %i)",
                               (readerTM == nullptr) ? "" : readerTM,
                               (identificationId == nullptr) ? "" : identificationId,
                               (identificationPin == nullptr) ? "" : identificationPin,
                               (statusText == nullptr) ? "" : statusText,
                               (statusTextLength == nullptr) ? 0 : *statusTextLength,
                               (errorNumber == nullptr) ? 0 : *errorNumber,
                               0,
                               channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );

        channelBridge->logEnd( "\t%s\t%s\t%s\t%s\t%i\t%i\t(%i, %i)", readerTM, identificationId, strlen( identificationPin ) ? "*****" : "", statusText, *statusTextLength, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int iRes = channelBridge->accessControlIdentification( l_device_address, masterId, readerTM, identificationId, identificationPin, statusText, (int&)*statusTextLength );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 6:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_ENTRANCE2_OFF, errorNumber );
                break;

            case 28:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
                break;

            case 29:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
                break;

            default:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%s\t%s\t%s\t%i\t%i\t(%i, %i)", readerTM, identificationId, strlen( identificationPin ) ? "*****" : "", statusText, *statusTextLength, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCAccessControlOnlineAction( int connectionNumber, int deviceAddress, int masterId, const char *readerTM, int mask, int type, int duration, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", masterId );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || masterId < 1 || masterId > 9999
        || readerTM == nullptr || mask == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );

        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t%i\t%i\t(%i, %i)", (readerTM == nullptr) ? "" : readerTM, mask, type, duration, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );

        channelBridge->logEnd( "\t%s\t0x%08X\t%i\t%i\t%i\t(%i, %i)", (readerTM == nullptr) ? "" : readerTM, mask, type, duration, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int iRes = channelBridge->accessControlOnlineAction( l_device_address, masterId, readerTM, mask, type, duration );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 6:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_ENTRANCE2_OFF, errorNumber );
                break;

            case 28:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
                break;

            case 29:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
                break;

            default:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t0x%08X\t%i\t%i\t%i\t(%i, %i)", (readerTM == nullptr) ? "" : readerTM, mask, type, duration, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCAccessControlKnobCommand( int connectionNumber, int deviceAddress, int knobCommand, int masterId, int knobNumber, const char *params, char *buffer, int *bufferSize, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t%i\t%i", knobCommand, masterId, knobNumber );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || (knobCommand != IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_TEACH && knobCommand != IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_DELETE && knobCommand != IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_LIST
        && knobCommand != IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_CHANGE_BATTERY_BEGIN && knobCommand != IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_CHANGE_BATTERY_END)
        || masterId < 1 || masterId > 9999
        || knobNumber < 0 || knobNumber > 15
        || (knobCommand == IV_VALUE_ACCESS_CONTROL_KNOB_COMMAND_LIST && (buffer == nullptr || bufferSize == nullptr)) )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );

        channelBridge->logEnd( "\t%s\t%s\t%i\t%i\t(%i, %i)", (params == nullptr) ? "" : params, (buffer == nullptr) ? "" : buffer, (bufferSize == nullptr) ? 0 : *bufferSize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );

        channelBridge->logEnd( "\t%s\t%s\t%i\t%i\t(%i, %i)", (params == nullptr) ? "" : params, (buffer == nullptr) ? "" : buffer, (bufferSize == nullptr) ? 0 : *bufferSize, (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int iRes = channelBridge->accessControlKnobCommand( l_device_address, knobCommand, masterId, knobNumber, params, reinterpret_cast<uint8_t *>( buffer ), bufferSize );
    if ( iRes == IV_ERROR_SUCCESS )
    {
        // Kein Fehler
        iRes = 1;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_UNAVAILABLE, errorNumber );
            }
            else
            {
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 6:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_ENTRANCE2_OFF, errorNumber );
                break;

            case 10: // Das anzusprechende Hardwaremodul (Zutrittsbus) ist nicht vorhanden.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_MODULE_ABSENCE, errorNumber );
                break;

            case 11: // Kein U&Z Funkmodul im Zutrittsbus vorhanden.
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NO_UUZ_MODULE, errorNumber );
               break;

            case 28:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_NEED_PASSWORD, errorNumber );
                break;

            case 29:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PASSWORD_LOCK, errorNumber );
                break;

            default:
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_PROCESSING, errorNumber );
            }
        }

        iRes = 0;
    }
    else
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes, errorNumber );
        iRes = 0;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%s\t%i\t%i\t(%i, %i)", (params == nullptr) ? "" : params, (buffer == nullptr) ? "" : buffer, (bufferSize == nullptr) ? 0 : *bufferSize, (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}
