#include "api/df_access_control_lists.h"
#include "datafox_library_iv.h"

#if defined(_MSC_VER) || defined(__MINGW32__)
#define snprintf	    _snprintf
#endif

DF_API_EXTERN_DECORATE void DF_API_CALL DFCClearEntrance2ListBuffer( int connectionNumber, int number )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return;
    }

    channelBridge->logBegin( __func__, "\t%i", number );

    // zulaessige Bereiche pruefen
    if ( number < -1 || number > 7 )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER );

        channelBridge->logEnd( "" );
        return;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS );

    channelBridge->ClearEntrance2ListBuffer( number );

    channelBridge->logEnd( "" );
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCMakeEntrance2List( int connectionNumber, int number, int recordCount, int size, const unsigned char *buffer, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( "%s\t%i\t%i\t%i\t%i", __func__, connectionNumber, number, recordCount, size );

    // zulaessige Bereiche pruefen
    if ( number < 0 || number > 7 || buffer == nullptr )
    {
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );

        channelBridge->logEnd( "\t\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( 0 ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( 0, IV_ERROR_SUCCESS, errorNumber );

    // ListNum ist:
    // 0: Reader         (33 Byte)
    // 1: Identification (64 Byte)
    // 2: Location       (20 Byte)
    // 3: Time           (25 Byte)
    // 4: Holiday        (21 Byte)
    // 5: Events         (17 Byte)
    // 6: Action         (21 Byte)
    // 7: Action2        (39 Byte)

    // Bereich der gueltigen Nummer pruefen.
    switch ( number )
    {
    case 0:
        // Anzahl moeglicher Datensaetze
        if ( (recordCount < 1) /*|| (iLineCount > (FLASH_SIZE_LIST / ENTRANCE2_RECORD_SIZE_0))*/ )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        // Berechnete Listengroesze pruefen
        if ( size != (recordCount * ENTRANCE2_RECORD_SIZE_0) )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        break;

    case 1:
        if ( (recordCount < 1) /*|| (iLineCount > (FLASH_SIZE_LIST / ENTRANCE2_RECORD_SIZE_1))*/ )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }

        if ( size != (recordCount * ENTRANCE2_RECORD_SIZE_1) )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        break;

    case 2:
        if ( (recordCount < 1) /*|| (iLineCount > (FLASH_SIZE_LIST / ENTRANCE2_RECORD_SIZE_2))*/ )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }

        if ( size != (recordCount * ENTRANCE2_RECORD_SIZE_2) )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        break;

    case 3:
        if ( (recordCount < 1) /*|| (iLineCount > (FLASH_SIZE_LIST / ENTRANCE2_RECORD_SIZE_3))*/ )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }

        if ( size != (recordCount * ENTRANCE2_RECORD_SIZE_3) )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        break;

    case 4:
        // Anzahl moeglicher Datensaetze
        if ( (recordCount < 1) /*|| (iLineCount > (FLASH_SIZE_LIST / ENTRANCE2_RECORD_SIZE_4))*/ )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        // Berechnete Listengroesze pruefen
        if ( size != (recordCount * ENTRANCE2_RECORD_SIZE_4) )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        break;

    case 5:
        // Anzahl moeglicher Datensaetze
        if ( (recordCount < 1) /*|| (iLineCount > (FLASH_SIZE_LIST / ENTRANCE2_RECORD_SIZE_5))*/ )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        // Berechnete Listengroesze pruefen
        if ( size != (recordCount * ENTRANCE2_RECORD_SIZE_5) )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        break;

    case 6:
        // Anzahl moeglicher Datensaetze
        if ( (recordCount < 1) /*|| (iLineCount > (FLASH_SIZE_LIST / ENTRANCE2_RECORD_SIZE_6))*/ )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        // Berechnete Listengroesze pruefen
        if ( size != (recordCount * ENTRANCE2_RECORD_SIZE_6) )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        break;

    case 7:
        // Anzahl moeglicher Datensaetze
        if ( (recordCount < 1) /*|| (iLineCount > (FLASH_SIZE_LIST / ENTRANCE2_RECORD_SIZE_7))*/ )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        // Berechnete Listengroesze pruefen
        if ( size != (recordCount * ENTRANCE2_RECORD_SIZE_7) )
        {
            channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
            break;
        }
        break;

    default:
        channelBridge->setLastErrorNumber( 0, IV_ERROR_FUNCTION_PARAMETER, errorNumber );
        break;
    }

    int iRes = 0;
    if ( channelBridge->getLastErrorNumber( 0 ) == IV_ERROR_SUCCESS )
    {
        iRes = channelBridge->MakeEntrance2List( number, recordCount, size, buffer );
    }

    channelBridge->logEnd( "\t\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( 0 ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCLoadEntrance2List( int connectionNumber, int deviceAddress, int number, int *errorNumber )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        if ( errorNumber )
        {
            *errorNumber = IV_ERROR_CHANNEL;
        }

        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", number );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || number < -1 || number > 7 || errorNumber == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER, errorNumber );

        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS, errorNumber );

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED, errorNumber );

        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Anderer Blocktransfer besteht bereits?
    if ( channelBridge->getBlockTransferState( l_device_address ) == ControllerSegment::BlockTransferState::active
        && channelBridge->getBlockTransferType( l_device_address ) != ControllerSegment::btt_access_control_list )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_BLOCK_TRANSFER_ACTIVE, errorNumber );

        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Zutrittskontrolle Version 2 wird erst ab 04.01.xx unterstuetzt.
    unsigned char protocolVersion = channelBridge->getProtocolVersion( l_device_address );
    if ( protocolVersion < 3 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_VERSION_FIRMWARE, errorNumber );

        channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    int iRes;
    /*
    char version[32];
    memset(version, 0, sizeof(version));
    // Firmwareversion lesen
    if ((iRes = channelBridge->getVersionFirmware(l_device_address, version)) != IV_ERROR_SUCCESS)
    {
        channelBridge->setLastErrorNumber(l_device_address, (iRes < 0) ? -iRes : IV_ERROR_VERSION_FIRMWARE, errorNumber);

        channelBridge->logEnd("\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber(l_device_address));
        return 0;
    }
    
    // Firmwareversion pruefen
    if (!theApp.m_util.CheckVersionFirmwareZK(version)) {

    *piError = IV_ERROR_FUNCTION_UNAVAILABLE;
    channelBridge->logEnd("\t%i\t(%i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0);
    return 0;
    }
    */
    // Listendaten wegschreiben. BIN
    if ( channelBridge->logIsWriteList()
         && channelBridge->getBlockTransferState( l_device_address ) == ControllerSegment::BlockTransferState::inactive )
    {
        char fileNameBin[_MAX_PATH];
        char basePath[340];
        int listNum;
        time_t t;
        tm *ltm;

        t = time( nullptr );
        ltm = localtime( &t );

        for ( listNum = 0; listNum < 8; listNum++ )
        {
            // Dateiname der Logdatei
            std::string logBasePath = channelBridge->logGetBasePath( basePath );
            int res = snprintf( fileNameBin, sizeof( fileNameBin ) - 1, LOG_LIST_FILENAME_BIN, logBasePath.c_str(),
                ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday,
                ltm->tm_hour, ltm->tm_min, ltm->tm_sec, connectionNumber, 'E', listNum );

            if ( res > static_cast<int>(sizeof( fileNameBin )) )
            {
                assert( false );
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FILE_OPEN, errorNumber );
                channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
                return 0;
            }

            // Daten in Datei schreiben.
            channelBridge->WriteListEntrance2ToBinFile( listNum, fileNameBin );
        }
    }

    // Listendaten wegschreiben. TXT
    if ( channelBridge->logIsWriteList() > 1
         && channelBridge->getBlockTransferState( l_device_address ) == ControllerSegment::BlockTransferState::inactive )
    {
        char fileNameTxt[_MAX_PATH];
        char basePath[340];
        std::string listName;
		unsigned char fieldCount, varList;
        int listNum;
        time_t t;
        tm *ltm;

        t = time( nullptr );
        ltm = localtime( &t );

        for ( listNum = 0; listNum < 8; listNum++ )
        {
            if ( !channelBridge->GetTableRecord( 200 + listNum, listName, &fieldCount, &varList, TRUE ) )
                break;

            // Dateiname der Logdatei
            std::string logBasePath = channelBridge->logGetBasePath( basePath );
            int res = snprintf( fileNameTxt, sizeof( fileNameTxt ) - 1, LOG_LIST_FILENAME_TXT, logBasePath.c_str(),
                                ltm->tm_year % 100, ltm->tm_mon + 1, ltm->tm_mday,
                                ltm->tm_hour, ltm->tm_min, ltm->tm_sec, connectionNumber, 'E', listName.c_str() );

            if ( res > static_cast<int>(sizeof( fileNameTxt )) )
            {
                assert( false );
                channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FILE_OPEN, errorNumber );
                channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, 0, channelBridge->getLastErrorNumber( l_device_address ) );
                return 0;
            }

            // Daten in die Datei schreiben.
            channelBridge->WriteListEntrance2ToTextFile( listNum, fileNameTxt );
        }
    }

    int l_error_number = IV_ERROR_SUCCESS;
    iRes = channelBridge->writeEntrance2ListData( l_device_address, number, &l_error_number );
    // Nach Kommunikationsfehler, Kanal bereinigen und ggf. Fehlernummer uebergeben.
    if ( iRes == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, l_error_number, errorNumber );
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );

        if ( channelBridge->getLastErrorNumber( l_device_address ) != IV_ERROR_BLOCK_TRANSFER_BREAK )
        {
            channelBridge->blockTransferDiscard( l_device_address );
        }
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", (errorNumber == nullptr) ? 0 : *errorNumber, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}
