#include "api/df_lists.h"
#include "datafox_library_iv.h"
#include <cstring>

DF_API_EXTERN int DF_API_CALL DFCTableOpen( int connectionNumber, int deviceAddress, const char *name, int options, int *tableHandle )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "" );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || name == nullptr || tableHandle == nullptr || options != 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", (name == nullptr) ? "" : name, options, (tableHandle == nullptr) ? 0 : *tableHandle, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", name, options, *tableHandle, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    std::string reserved;
    int l_table_handle;
    int iRes = channelBridge->tableOpen( l_device_address, name, static_cast<uint32_t>( options ), l_table_handle, reserved );
    if ( iRes > IV_ERROR_SUCCESS )
    { 
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        *tableHandle = l_table_handle;
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t%i\t(%i, %i)", name, options, *tableHandle, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCOpenTable( int connectionNumber, int deviceAddress, const char *name, int options, int *tableHandle, const char *reserved, int *errorNumber )
{
    (void)reserved;
    int res = DFCTableOpen( connectionNumber, deviceAddress, name, options, tableHandle );

    if ( !res && errorNumber != nullptr)
    {
        ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
        // Wurde Channel-Objekt ermittelt.
        if ( channelBridge != nullptr )
        {
            *errorNumber = channelBridge->getLastErrorNumber( channelBridge->cast_device_address( deviceAddress ) );
        }
    }

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCTableClose( int connectionNumber, int deviceAddress, int tableHandle )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", tableHandle ); 

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    int iRes = channelBridge->tableClose( l_device_address, tableHandle );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    { 
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCCloseTable( int connectionNumber, int deviceAddress, int tableHandle, int *errorNumber )
{
    int res = DFCTableClose( connectionNumber, deviceAddress, tableHandle );

    if ( !res && errorNumber != nullptr )
    {
        ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
        // Wurde Channel-Objekt ermittelt.
        if ( channelBridge != nullptr )
        {
            *errorNumber = channelBridge->getLastErrorNumber( channelBridge->cast_device_address( deviceAddress ) );
        }
    }

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCTableSetFilter( int connectionNumber, int deviceAddress, int tableHandle, const char *value )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", tableHandle );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || value == nullptr || std::strlen( value ) < 3) // mindestens Filter "a=b"
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%s\t(%i, %i)", (value == nullptr) ? "" : value, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%s\t(%i, %i)", value, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    int iRes = channelBridge->tableSetFilter( l_device_address, tableHandle, value );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 3:
                l_error_number = IV_ERROR_NULL_RESULT;
                break;

            case 4:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t(%i, %i)", value, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetFilter( int connectionNumber, int deviceAddress, int tableHandle, const char *value, int *errorNumber )
{
    int res = DFCTableSetFilter( connectionNumber, deviceAddress, tableHandle, value );

    if ( !res && errorNumber != nullptr )
    {
        ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
        // Wurde Channel-Objekt ermittelt.
        if ( channelBridge != nullptr )
        {
            *errorNumber = channelBridge->getLastErrorNumber( channelBridge->cast_device_address( deviceAddress ) );
        }
    }

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCTableGetFilter( int connectionNumber, int deviceAddress, int tableHandle, char *value, int *valueSize )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }


    channelBridge->logBegin( __func__, deviceAddress, "\t%i", tableHandle );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || value == nullptr || valueSize == nullptr || *valueSize < 2 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (value == nullptr) ? "" : value, (valueSize == nullptr) ? 0 : *valueSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", value, *valueSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    std::string lValue;
    int iRes = channelBridge->tableGetFilter( l_device_address, tableHandle, lValue );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        // Zu kleinen Puffer abfangen!
        if ( *valueSize < static_cast<int>(lValue.size() + 1) )
        {
            channelBridge->setLastErrorNumber( l_device_address, -IV_ERROR_FUNCTION_PARAMETER );
            iRes = 0;
        }
        else
        {
            *valueSize = static_cast<int>(lValue.copy( value, *valueSize - 1 ));
            value[*valueSize] = 0; // Nullterminierung
            iRes = 1;
        }
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", value, *valueSize, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetFilter( int connectionNumber, int deviceAddress, int tableHandle, char *value, int *valueLength, int *errorNumber )
{
    int res = DFCTableGetFilter( connectionNumber, deviceAddress, tableHandle, value, valueLength );

    if ( !res && errorNumber != nullptr )
    {
        ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
        // Wurde Channel-Objekt ermittelt.
        if ( channelBridge != nullptr )
        {
            *errorNumber = channelBridge->getLastErrorNumber( channelBridge->cast_device_address( deviceAddress ) );
        }
    }

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCTableRemoveFilter( int connectionNumber, int deviceAddress, int tableHandle )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", tableHandle );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    int iRes = channelBridge->tableRemoveFilter( l_device_address, tableHandle );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCClearFilter( int connectionNumber, int deviceAddress, int tableHandle, int *errorNumber )
{
    int res = DFCTableRemoveFilter( connectionNumber, deviceAddress, tableHandle );

    if ( !res && errorNumber != nullptr )
    {
        ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
        // Wurde Channel-Objekt ermittelt.
        if ( channelBridge != nullptr )
        {
            *errorNumber = channelBridge->getLastErrorNumber( channelBridge->cast_device_address( deviceAddress ) );
        }
    }

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCTableGetRowCount( int connectionNumber, int deviceAddress, int tableHandle, int *currentCount, int *unsortedCount, int *deletedCount )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", tableHandle );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || currentCount == nullptr || unsortedCount == nullptr || deletedCount == nullptr )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%i\t%i\t%i\t(%i, %i)", (currentCount == nullptr) ? 0 : *currentCount, (unsortedCount == nullptr) ? 0 : *unsortedCount, (deletedCount == nullptr) ? 0 : *deletedCount, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%i\t%i\t%i\t(%i, %i)", *currentCount, *unsortedCount, *deletedCount, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    uint32_t l_current_count;
    uint32_t l_unsorted_count;
    uint32_t l_deleted_count;
    int iRes = channelBridge->tableGetRowCount( l_device_address, tableHandle, l_current_count, l_unsorted_count, l_deleted_count );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 3:
                l_error_number = IV_ERROR_VALUE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes > IV_ERROR_SUCCESS )
    { 
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        *currentCount = static_cast<int>( l_current_count );
        *unsortedCount = static_cast<int>( l_unsorted_count );
        *deletedCount = static_cast<int>( l_deleted_count );
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t%i\t%i\t(%i, %i)", *currentCount, *unsortedCount, *deletedCount, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN int DF_API_CALL DFCTableGetCurrentRow( int connectionNumber, int deviceAddress, int tableHandle, int *currentRow )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i", tableHandle );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || currentRow == nullptr)
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%i\t(%i, %i)", (currentRow == nullptr) ? 0 : *currentRow, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%i\t(%i, %i)", *currentRow, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    uint32_t l_current_row;
    int iRes = channelBridge->tableGetCurrentRow( l_device_address, tableHandle, l_current_row );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 3:
                l_error_number = IV_ERROR_VALUE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    { 
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        *currentRow = static_cast<int>( l_current_row );
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%i\t(%i, %i)", *currentRow, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN int DF_API_CALL DFCTableSetCurrentRow( int connectionNumber, int deviceAddress, int tableHandle, int offset, int direction )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t%i\t%i", tableHandle, offset, direction );  

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || (direction != 0 && direction != 1 && direction != 2) )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    int iRes = channelBridge->tableSetCurrentRow( l_device_address, tableHandle, offset, static_cast<uint16_t>( direction ) );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 3:
                l_error_number = IV_ERROR_BOF;
                break;

            case 4:
                l_error_number = IV_ERROR_EOF;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSkip( int connectionNumber, int deviceAddress, int tableHandle, int offset, int origin, int *errorNumber )
{
    int res = DFCTableSetCurrentRow( connectionNumber, deviceAddress, tableHandle, offset, origin );

    if ( !res && errorNumber != nullptr )
    {
        ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
        // Wurde Channel-Objekt ermittelt.
        if ( channelBridge != nullptr )
        {
            *errorNumber = channelBridge->getLastErrorNumber( channelBridge->cast_device_address( deviceAddress ) );
        }
    }

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCTableSetCurrentRowData( int connectionNumber, int deviceAddress, int tableHandle, int flags, char separator, const char *data, int dataSize )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t0x%08X\t%c", tableHandle, flags, separator );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || flags < 0 || flags > 1
        || (separator != 0 && separator != '\t' && separator != ';') || data == nullptr || (separator == 0 && dataSize < 2) )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );

        if (separator == 0 && data != nullptr)
        {

        }
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (data == nullptr) ? "" : data, dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", data, dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    if ( dataSize > 0xFFFF )
    {
        // Wird bei der Ausfhrung abhngig vom Protokoll ggf. einen Parameterfehler 20 ergeben, 
        // da eine solch groe Menge nicht gltig ist.
        dataSize = 0xFFFF;
    }

    int iRes = channelBridge->tableSetCurrentRowData( l_device_address, tableHandle, static_cast<uint16_t>( flags ), separator, data, static_cast<uint16_t>( dataSize ) );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 3:
                l_error_number = IV_ERROR_VALUE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    { 
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", data, dataSize, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN int DF_API_CALL DFCTableSetCurrentColumnData( int connectionNumber, int deviceAddress, int tableHandle, int flags, const char *name, const char *data, int dataSize )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t0x%08X", tableHandle, flags );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || flags < 0 || flags > 1
        || name == nullptr || data == nullptr 
        || (dataSize == 0 && std::strlen( data ) == 0) 
        || (dataSize == 0 && std::strlen( data ) > 250)
        || dataSize > 440 ) // Binaerfelder koennen als Hex-String geschickt werden.
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%s\t%s\t%i\t(%i, %i)", (name == nullptr) ? "" : name, (data == nullptr) ? "" : data, dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%s\t%s\t%i\t(%i, %i)", name, data, dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    uint16_t l_data_size = (dataSize > 0xFFFF) ? 0xFFFF : static_cast<uint16_t>( dataSize );
    if ( l_data_size == 0 )
    {
        l_data_size = static_cast<uint16_t>( std::strlen( data ) );
    }

    int iRes = channelBridge->tableSetCurrentColumnData( l_device_address, tableHandle, static_cast<uint16_t>( flags ), name, data, l_data_size );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 3:
                l_error_number = IV_ERROR_VALUE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    { 
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%s\t%i\t(%i, %i)", name, data, dataSize, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCSetField( int connectionNumber, int deviceAddress, int tableHandle, const char *name, const char *value, int *errorNumber )
{
    int dataSize = static_cast<int>( strlen( value ) );
    int res = DFCTableSetCurrentColumnData( connectionNumber, deviceAddress, tableHandle, 0, name, value, dataSize );

    if ( !res && errorNumber != nullptr )
    {
        ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
        // Wurde Channel-Objekt ermittelt.
        if ( channelBridge != nullptr )
        {
            *errorNumber = channelBridge->getLastErrorNumber( channelBridge->cast_device_address( deviceAddress ) );
        }
    }

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCTableSetAllRowsToColumnData( int connectionNumber, int deviceAddress, int tableHandle, int flags, const char *name, const char *data, int dataSize )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t0x%08X", tableHandle, flags );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || flags < 0 || flags > 0
        || name == nullptr || data == nullptr 
        || (dataSize == 0 && std::strlen( data ) == 0) 
        || (dataSize == 0 && std::strlen( data ) > 250)
        || dataSize > 440 ) // Binaerfelder koennen als Hex-String geschickt werden.
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%s\t%s\t%i\t(%i, %i)", (name == nullptr) ? "" : name, (data == nullptr) ? "" : data, dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%s\t%s\t%i\t(%i, %i)", name, data, dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    uint16_t l_data_size = (dataSize > 0xFFFF) ? 0xFFFF : static_cast<uint16_t>( dataSize );
    if ( l_data_size == 0 )
    {
        l_data_size = static_cast<uint16_t>( std::strlen( data ) );
    }

    int iRes = channelBridge->tableSetAllRowsToColumnData( l_device_address, tableHandle, static_cast<uint16_t>( flags ), name, data, l_data_size );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 3:
                l_error_number = IV_ERROR_VALUE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    { 
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%s\t%i\t(%i, %i)", name, data, dataSize, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN int DF_API_CALL DFCTableGetCurrentRowData( int connectionNumber, int deviceAddress, int tableHandle, int flags, char separator, char *data, int *dataSize )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t0x%08X\t%c", tableHandle, flags, separator );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || flags < 0 || flags > 1
        || (separator != 0 && separator != '\t' && separator != ';') || data == nullptr || dataSize == nullptr || *dataSize < 2 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );

        if (separator == 0 && data != nullptr)
        {

        }
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (data == nullptr) ? "" : data, (dataSize == nullptr) ? 0 : *dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", data, *dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    uint16_t l_data_size = (*dataSize > 0xFFFF) ? 0xFFFF : static_cast<uint16_t>( *dataSize );
    int iRes = channelBridge->tableGetCurrentRowData( l_device_address, tableHandle, static_cast<uint16_t>( flags ), separator, data, l_data_size );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 3:
                l_error_number = IV_ERROR_VALUE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    { 
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        *dataSize = l_data_size;
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", data, *dataSize, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN int DF_API_CALL DFCTableGetCurrentColumnData( int connectionNumber, int deviceAddress, int tableHandle, int flags, const char *name, char *data, int *dataSize )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t0x%08X", tableHandle, flags );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || flags < 0 || flags > 1
        || name == nullptr || data == nullptr || dataSize == nullptr || *dataSize < 2 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t%s\t%s\t%i\t(%i, %i)", (name == nullptr) ? "" : name, (data == nullptr) ? "" : data, (dataSize == nullptr) ? 0 : *dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%s\t%s\t%i\t(%i, %i)", name, data, *dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    uint16_t l_data_size = (*dataSize > 0xFFFF) ? 0xFFFF : static_cast<uint16_t>( *dataSize );
    int iRes = channelBridge->tableGetCurrentColumnData( l_device_address, tableHandle, static_cast<uint16_t>( flags ), name, data, l_data_size );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    {
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        *dataSize = l_data_size;
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%s\t%i\t(%i, %i)", name, data, *dataSize, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN_DECORATE int DF_API_CALL DFCGetField( int connectionNumber, int deviceAddress, int tableHandle, const char *name, char *value, int *valueLength, int *errorNumber )
{
    int res = DFCTableGetCurrentColumnData( connectionNumber, deviceAddress, tableHandle, 0, name, value, valueLength );

    if ( !res && errorNumber != nullptr )
    {
        ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
        // Wurde Channel-Objekt ermittelt.
        if ( channelBridge != nullptr )
        {
            *errorNumber = channelBridge->getLastErrorNumber( channelBridge->cast_device_address( deviceAddress ) );
        }
    }

    return res;
}

DF_API_EXTERN int DF_API_CALL DFCTableAppendRowData( int connectionNumber, int deviceAddress, int tableHandle, int flags, char separator, const char *data, int dataSize )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t0x%08X\t%c", tableHandle, flags, separator );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0
        || flags < 0 || flags > 0
        || (separator != 0 && separator != '\t' && separator != ';') || data == nullptr || (separator == 0 && dataSize < 2) )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );

        if (separator == 0 && data != nullptr)
        {

        }
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", (data == nullptr) ? "" : data, dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", data, dataSize, 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    if ( dataSize > 0xFFFF )
    {
        // Wird je nach Protokoll ggf. ein Parameterfehler ergeben.
        // zu groe Datenmenge.
        dataSize = 0xFFFF;
    }

    int iRes = channelBridge->tableAppendRowData( l_device_address, tableHandle, static_cast<uint16_t>( flags ), separator, data, static_cast<uint16_t>( dataSize ) );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 3:
                l_error_number = IV_ERROR_VALUE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    { 
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t%s\t%i\t(%i, %i)", data, dataSize, iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN int DF_API_CALL DFCTableDeleteCurrentRow( int connectionNumber, int deviceAddress, int tableHandle, int flags )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t0x%08X", tableHandle, flags );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || flags < 0 || flags > 1 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS);

    int iRes = channelBridge->tableDeleteCurrentRow( l_device_address, tableHandle, static_cast<uint16_t>( flags ) );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 3:
                l_error_number = IV_ERROR_VALUE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    { 
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}

DF_API_EXTERN int DF_API_CALL DFCTableDeleteAvailableRows( int connectionNumber, int deviceAddress, int tableHandle, int flags )
{
    ChannelBridge *channelBridge = theApp.channelBridge( connectionNumber );
    // Wurde Channel-Objekt ermittelt.
    if ( channelBridge == nullptr )
    {
        return 0;
    }

    channelBridge->logBegin( __func__, deviceAddress, "\t%i\t0x%08X", tableHandle, flags );

    // Gerteadresse auf lokale mappen.
    uint8_t l_device_address = channelBridge->cast_device_address( deviceAddress );

    // zulaessige Bereiche pruefen
    if ( l_device_address == 0 || flags < 0 || flags > 1 )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_FUNCTION_PARAMETER );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Schnittstelle geoeffnet?
    if ( channelBridge->isOpen() == false )
    {
        channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_CHANNEL_CLOSED );
        channelBridge->logEnd( "\t(%i, %i)", 0, channelBridge->getLastErrorNumber( l_device_address ) );
        return 0;
    }

    // Fehlerwert zuruecksetzen.
    channelBridge->setLastErrorNumber( l_device_address, IV_ERROR_SUCCESS );

    int iRes = channelBridge->tableDeleteAvailableRows( l_device_address, tableHandle, static_cast<uint16_t>( flags ) );
    if ( iRes > IV_ERROR_SUCCESS )
    {
        int l_error_number = IV_ERROR_SUCCESS;
        // Fehler bei Durchfuehrung des Befehls
        // Modulfehler auswerten.
        if ( (iRes >> 8) & 0xFF )
        {
            if ( ((iRes >> 8) & 0xFF) == 2 )
            {
                l_error_number = IV_ERROR_FUNCTION_UNAVAILABLE;
            }
            else
            {
                l_error_number = IV_ERROR_PROCESSING;
            }
        }
        else
        {
            // Funktionsfehler auswerten.
            switch ( iRes & 0xFF )
            {
            case 1:
                l_error_number = IV_ERROR_HANDLE;
                break;

            case 2:
                l_error_number = IV_ERROR_NAME_UNKNOWN;
                break;

            case 3:
                l_error_number = IV_ERROR_VALUE;
                break;

            case 28:
                l_error_number = IV_ERROR_NEED_PASSWORD;
                break;

            case 29:
                l_error_number = IV_ERROR_PASSWORD_LOCK;
                break;

            default:
                l_error_number = IV_ERROR_PROCESSING;
            }
        }

        channelBridge->setLastErrorNumber( l_device_address, l_error_number );
        iRes = 0;
    }
    else if ( iRes < IV_ERROR_SUCCESS )
    { 
        // Fehler waehrend Kommunikation oder Verarbeitung in DLL
        channelBridge->setLastErrorNumber( l_device_address, -iRes );
        iRes = 0;
    }
    else
    {
        iRes = 1;
    }

    // Nach Kommunikationsfehler, Kanal bereinigen
    if ( iRes == 0 )
    {
        channelBridge->resetWithError( channelBridge->getLastErrorNumber( l_device_address ) );
    }

    channelBridge->logEnd( "\t(%i, %i)", iRes, channelBridge->getLastErrorNumber( l_device_address ) );

    return iRes;
}
