#pragma once
#ifndef DEVICE_INFO_H
#define DEVICE_INFO_H

#include "hardware_info.h"
#include "software_info.h"
#include "firmware_options.h"
#include <cstdint>

class DeviceInfo
{
public:
    DeviceInfo();
    virtual ~DeviceInfo();

public:
    void clear();

    void setVersionBootloader( const std::string &version );
    std::string versionBootloader() const;

    void setBootloaderMode( bool active );
    bool bootloaderMode() const;

    void setHardwareInfo( const HardwareInfo &hardwareInfo );
    HardwareInfo &hardwareInfo();
    
    void setFileHardwareInfo( const FileHardwareInfo &file );
    FileHardwareInfo &fileHardwareInfo();
    listItemBoard collectModuleIds( const setModuleId &ids );

    void setVersionFirmware( const std::string &version );
    std::string versionFirmware() const;

    void setSoftwareInfo( const SoftwareInfo &softwareInfo );
    SoftwareInfo &softwareInfo();

    void setFirmwareOptions( const FirmwareOptions &firmwareOptions );
    FirmwareOptions &firmwareOptions();

    void setSerialNumber( uint32_t serial );
    uint32_t serialNumber() const;

    void setPasswordKey( const std::string &key );
    std::string passwordKey() const;

    void setVersionFingerprint( const std::string &version );
    std::string versionFingerprint() const;

    void setVersionUSBHost( const std::string &version );
    std::string versionUSBHost() const;

    void setVersionGPS( const std::string &version );
    std::string versionGPS() const;

    void setVersionTCPIP( const std::string &version );
    std::string versionTCPIP() const;

    void setMacAddress( const std::string &address );
    std::string macAddress() const;

    void setAccessControlMasterId( const std::string &id );
    std::string accessControlMasterId() const;

    void setTimeboyGroupId( const std::string &id );
    std::string timeboyGroupId() const;

    void setBoardApplication( const std::string &val );
    std::string boardApplication() const;

    void setBoardPlace_0_Output_1_Mode( const std::string &mode );
    std::string boardPlace_0_Output_1_Mode() const;

    void setBoardPlace_0_Serial_1_Use( const std::string &use );
    std::string boardPlace_0_Serial_1_Use() const;

    void setBoardPlace_0_Serial_2_Use( const std::string &use );
    std::string boardPlace_0_Serial_2_Use() const;

    void setTimeStartOS( const std::string &val );
    std::string timeStartOS() const;

    void setOffTimeLong( const std::string &val );
    std::string offTimeLong() const;

    void setOffTimeShort( const std::string &val );
    std::string offTimeShort() const;

    void setOnTime( const std::string &val );
    std::string onTime() const;

    void set_restore_ac_power_loss( const std::string &val );
    std::string get_restore_ac_power_loss() const;

    void setAccessModulesState( const std::string &val );
    std::string getAccessModulesState() const;

public: // Helper / Shorties
    bool importStructHardwareInfo( const StructHardwareInfo &info );
    StructHardwareInfo structHardwareInfo();

    void setStructSoftwareInfo( const StructSoftwareInfo &info );
    StructSoftwareInfo structSoftwareInfo() const;

    StructSoftwareInfo checkInfo( const StructSoftwareInfo &info );
    int SoftwareInfoToString( const StructSoftwareInfo &swInfo, std::string &info );

    int compareInfos( const StructHardwareInfo &hwInfo, const StructSoftwareInfo &swInfo, const std::string &versionFirmware, bool checkBoardRevision = true ); 

public: // Shorties für FirmwareOptions
    void setSystemFlags( uint32_t flags );
    uint32_t systemFlags() const;

    void setTypeExternalReader( uint8_t type );
    uint8_t getTypeExternalReader();

    void setOptionIOModule( uint8_t enabled );
    uint8_t getOptionIOModule();

    void setOptionTBAdaption( uint8_t enabled );
    uint8_t getOptionTBAdaption();

    void setOptionComServer( uint8_t enabled );
    uint8_t getOptionComServer();

    void setOptionTPWrite( uint8_t enabled );
    uint8_t getOptionTPWrite();

    void setOptionHTTP( uint8_t enabled );
    uint8_t getOptionHTTP();

    void setOptionLanguage( uint8_t enabled );
    uint8_t getOptionLanguage();

    void setOptionBiokeyUpdate( uint8_t enabled );
    uint8_t getOptionBiokeyUpdate();

    void setSetupOptions( uint16_t mask, uint16_t options );
    uint16_t getSetupOptions( uint16_t mask );

    void setPayableOptions( uint8_t mask, uint8_t options );
    bool isPayableOption( FirmwareOptions::EnumPayableOption option );

    void setMainInterfaceType( uint8_t val )
    {
        m_mainInterfaceType = val;
    }
    uint8_t mainInterfaceType() const
    {
        return m_mainInterfaceType;
    }

public: // Fehlerbehandlung
    std::string m_errorText;

private: // Variablen
    std::string m_versionBootloader;
    bool m_bootloaderMode;
    std::string m_versionFirmware;

    uint32_t m_serialNumber;

    HardwareInfo m_hardwareInfo;

    std::string m_versionFingerprint;
    std::string m_versionUSBHost;
    std::string m_versionGPS;
    std::string m_versionTCPIP;

    std::string m_macAddress;

    SoftwareInfo m_softwareInfo;
    std::string m_passwordKey;

    FirmwareOptions m_firmwareOptions;

    std::string m_accessControlMasterId;
    std::string m_timeboyGroupId;

    std::string m_boardApplication;
    std::string m_boardPlace_0_Output_1_Mode;
    std::string m_boardPlace_0_Serial_1_Use;
    std::string m_boardPlace_0_Serial_2_Use;

    std::string m_timeStartOS;
    std::string m_offTimeLong;
    std::string m_offTimeShort;
    std::string m_onTime;
    std::string m_restoreAcPowerLoss;

    std::string m_accessModulesState;

    uint8_t m_mainInterfaceType;
};

#endif
